# Copyright (c) ONNX Project Contributors
# SPDX-License-Identifier: Apache-2.0
"""Shape inference operators.

This module imports all operator shape inference functions to ensure
they are registered with the global registry.
"""

# Import to trigger registration
from onnx_shape_inference._ops import (
    _ai_onnx_ml,
    _arg,
    _attention,
    _cast,
    _concat,
    _constant,
    _constant_of_shape,
    _control_flow,
    _conv,
    _conv_transpose,
    _data_dependent,
    _depth_space,
    _dropout,
    _einsum,
    _elementwise,
    _expand,
    _gather,
    _gemm,
    _loss,
    _matmul,
    _microsoft,
    _misc,
    _normalization,
    _onehot,
    _optional,
    _pad,
    _pooling,
    _random,
    _range,
    _reduce,
    _reshape,
    _resize,
    _rnn,
    _scatter,
    _sequence,
    _shape_ops,
    _signal,
    _slice,
    _softmax,
    _spatial,
    _split,
    _squeeze,
    _string,
    _tile,
    _topk,
    _training,
    _transpose,
    _unary,
    _where,
    _window,
)

__all__ = [
    "_ai_onnx_ml",
    "_arg",
    "_attention",
    "_microsoft",
    "_cast",
    "_concat",
    "_constant",
    "_constant_of_shape",
    "_control_flow",
    "_conv",
    "_conv_transpose",
    "_data_dependent",
    "_depth_space",
    "_dropout",
    "_einsum",
    "_elementwise",
    "_expand",
    "_gather",
    "_gemm",
    "_loss",
    "_matmul",
    "_misc",
    "_normalization",
    "_onehot",
    "_optional",
    "_pad",
    "_pooling",
    "_random",
    "_range",
    "_reduce",
    "_reshape",
    "_resize",
    "_rnn",
    "_scatter",
    "_sequence",
    "_shape_ops",
    "_signal",
    "_slice",
    "_softmax",
    "_spatial",
    "_split",
    "_squeeze",
    "_string",
    "_tile",
    "_topk",
    "_training",
    "_transpose",
    "_unary",
    "_where",
    "_window",
]
